local midi_col = {
	[0x0] = 0xe0e0e0ff,

	[0x8] = 0xe00040ff,
	[0x9] = 0x00e040ff,
	[0xa] = 0xe04000ff,
	[0xb] = 0xe0e000ff,
	[0xc] = 0x4040e0ff,
	[0xd] = 0xe040e0ff,
	[0xe] = 0x40e0e0ff,
	[0xf] = 0xe0e0e0ff,
}

local oscp = 'http://open-music-kontrollers.ch/lv2/osc#'
local OSC = {
	Message = Map(oscp .. 'Message'),
	messagePath = Map(oscp .. 'messagePath'),
	messageArguments = Map(oscp .. 'messageArguments'),
	Timestamp = Map(oscp .. 'Timestamp'),
	timestampIntegral = Map(oscp .. 'timestampIntegral'),
	timestampFraction = Map(oscp .. 'timestampFraction'),
	Nil = Map(oscp .. 'Nil'),
	Impulse = Map(oscp .. 'Impulse'),
	Char = Map(oscp .. 'Char'),
	RGBA = Map(oscp .. 'RGBA')
}

local midip = 'http://lv2plug.in/ns/ext/midi#'
local MIDI = {
	MidiEvent = Map(midip .. 'MidiEvent')
}

local atomp = 'http://lv2plug.in/ns/ext/atom#'
local ATOM = {
	Int = Map(atomp .. 'Int'),
	Long = Map(atomp .. 'Long'),
	Float = Map(atomp .. 'Float'),
	Double = Map(atomp .. 'Double'),
	Bool = Map(atomp .. 'Bool'),
	URID = Map(atomp .. 'URID'),
	String = Map(atomp .. 'String'),
	Literal = Map(atomp .. 'Literal'),
	Chunk = Map(atomp .. 'Chunk'),
	Object  = Map(atomp .. 'Object')
}

--[[
-- MIDI hexadecimal editing mode
--]]

local function midi_hex_parse(forge, str)
	local tab = {}
	for word in string.gmatch(str, '%s*(%x%x)') do
		local byte = tonumber(word, 16) -- base 16
		table.insert(tab, byte)
	end

	local sz = #tab
	if sz > 0 then
		local sys = tab[1]
		local cmd = sys >> 4

		-- check for valid size
		if   ((cmd == 0x8)  and (sz == 3)) -- note-off
			or ((cmd == 0x9)  and (sz == 3)) -- note-on
			or ((cmd == 0xa)  and (sz == 3)) -- note-pressure
			or ((cmd == 0xb)  and (sz == 3)) -- controller
			or ((cmd == 0xc)  and (sz == 3)) -- program-change
			or ((cmd == 0xd)  and (sz == 2)) -- channel-pressure
			or ((cmd == 0xe)  and (sz == 3)) -- pitch-bend

			or ((sys == 0xf1) and (sz == 2)) -- mtc-quarter
			or ((sys == 0xf2) and (sz == 3)) -- song-position
			or ((sys == 0xf3) and (sz == 2)) -- song-select
			or ((sys == 0xf6) and (sz == 1)) -- tune-request
			or ((sys == 0xf8) and (sz == 1)) -- clock
			or ((sys == 0xfa) and (sz == 1)) -- start
			or ((sys == 0xfb) and (sz == 1)) -- continue
			or ((sys == 0xfc) and (sz == 1)) -- stop
			or ((sys == 0xfe) and (sz == 1)) -- active-sensing
			or ((sys == 0xff) and (sz == 1)) -- reset
		then
			forge:midi(tab)
		end
	end
end

local function midi_hex_print(atom)
	local body = atom.body
	local tab = { string.byte(body, 1, #body) }
	for i, v in ipairs(tab) do
		tab[i] = string.format('%02X', v)
	end

	local res = table.concat(tab, ' ')
	local col = midi_col[string.byte(body, 1) >> 4]

	return res, col
end

local function midi_hex_edit(atom)
	if not atom or atom.type ~= MIDI.MidiEvent then
		return ''
	end

	return midi_hex_print(atom)
end

local function midi_hex_write(atom)
	if not atom then
		return '-- -- --', midi_col[0x0]
	elseif atom.type ~= MIDI.MidiEvent then
		return Unmap(atom.otype or atom.type), midi_col[0x0]
	end

	return midi_hex_print(atom)
end

local midi_hex_mode = {
	label = 'MIDI-Hex',
	parse= midi_hex_parse,
	edit= midi_hex_edit,
	write = midi_hex_write
}

--[[
-- MIDI decimal editing mode
--]]

local midi_offset = {
	['C'] = 0, ['C#'] = 1,
	['D'] = 2, ['D#'] = 3,
	['E'] = 4,
	['F'] = 5, ['F#'] = 6,
	['G'] = 7, ['G#'] = 8,
	['A'] = 9, ['A#'] = 10,
	['B'] = 11
}

local offset_midi = {
	[0]  = 'C ', [1]  = 'C#',
	[2]  = 'D ', [3]  = 'D#',
	[4]  = 'E ',
	[5]  = 'F ', [6]  = 'F#',
	[7]  = 'G ', [8]  = 'G#',
	[9]  = 'A ', [10] = 'A#',
	[11] = 'B '
}

local function midi_dec_parse(forge, str)
	local cmd, chan, offset = string.match(str, '%s*([NABCGP])(%d*)%s+()')
	chan = chan and tonumber(chan) or 1
	chan = chan and chan - 1 or 0

	if cmd == 'N' or cmd == 'A' then -- 0x90, 0x80, 0xa0
		local name, oct, msb = string.match(str, '([CDEFGAB]#?)%s*([+-]?%d?)%s*(%x*)', offset)
		oct = oct and tonumber(oct) or 4
		msb = msb and tonumber(msb) or 0
		local note = midi_offset[name] + (oct+1)*12
		cmd = cmd == 'N' and (msb == 0 and 0x80 or 0x90) or 0xa0
		forge:midi(cmd | chan, note, msb)
	elseif cmd == 'C' or cmd == 'G' then -- 0xb0, 0xc0
		local lsb, msb = string.match(str, '(%d+)%s+(%d+)', offset)
		lsb = lsb and tonumber(lsb) or 0
		msb = msb and tonumber(msb) or 0
		cmd = cmd == 'C' and 0xb0 or 0xc0
		forge:midi(cmd | chan, lsb, msb)
	elseif cmd == 'P' then -- 0xd0
		local lsb = string.match(str, '(%d+)', offset)
		lsb = lsb and tonumber(lsb) or 0
		cmd = 0xd0
		forge:midi(cmd | chan, lsb)
	elseif cmd == 'B' then -- 0xe0
		local bend = string.match(str, '([+-]?%d+)', offset)
		bend = bend and tonumber(bend) or 0
		local absolute = bend + 0x2000
		local lsb = absolute & 0x7f
		local msb = absolute >> 7
		cmd = 0xe0
		forge:midi(cmd | chan, lsb, msb)
	end
end

local function midi_dec_print(atom)
	local body = atom.body
	local sz = #body
	local byte = string.byte(body, 1)
	local cmd = byte >> 4
	local cha = (byte & 0xf) + 1
	local col = midi_col[cmd]

	if ( (cmd == 0x8) or (cmd == 0x9) ) and (sz == 3)  then
		local note = string.byte(body, 2)
		local msb = string.byte(body, 3)
		local off = note % 12
		local oct = (note // 12) - 1
		return string.format('N%02i %s %+2i %3i', cha, offset_midi[off], oct, msb), col
	elseif (cmd == 0xa) and (sz == 3) then
		local note = string.byte(body, 2)
		local msb = string.byte(body, 3)
		local off = note % 12
		local oct = (note // 12) - 1
		return string.format('A%02i %s %+2i %3i', cha, offset_midi[off], oct, msb), col
	elseif (cmd == 0xb) and (sz == 3) then
		local lsb = string.byte(body, 2)
		local msb = string.byte(body, 3)
		return string.format('C%02i %5i %3i', cha, lsb, msb), col
	elseif (cmd == 0xc) and (sz == 3) then
		local lsb = string.byte(body, 2)
		local msb = string.byte(body, 3)
		return string.format('G%02i %5i %3i', cha, lsb, msb), col
	elseif (cmd == 0xd) and (sz == 2) then
		local lsb = string.byte(body, 2)
		return string.format('P%02i %9i', cha, lsb), col
	elseif (cmd == 0xe) and (sz == 3) then
		local lsb = string.byte(body, 2)
		local msb = string.byte(body, 3)
		local bend = ( (msb << 7) | lsb) - 0x2000
		return string.format('B%02i %+9i', cha, bend), col
	else
		return '??? ?? ?? ???', col
	end
end

local function midi_dec_edit(atom)
	if not atom or atom.type ~= MIDI.MidiEvent then
		return ''
	end

	return midi_dec_print(atom)
end

local function midi_dec_write(atom)
	if not atom then
		return '--- -- -- ---', midi_col[0x0]
	elseif atom.type ~= MIDI.MidiEvent then
		return Unmap(atom.otype or atom.type), midi_col[0x0]
	end

	return midi_dec_print(atom)
end

local midi_dec_mode = {
	label = 'MIDI-Dec',
	parse= midi_dec_parse,
	edit= midi_dec_edit,
	write = midi_dec_write
}

--[[
-- OSC editing mode
--]]

local function osc_parse(forge, str)
	local path, offset = string.match(str, '%s*(/[^# ]*)()')

	if path then
		local fmt, offset = string.match(str, '%s*,?([ifsbTFNIhdtScrm]*)()', offset)

		if fmt then
			local tab = {}

			for c in string.gmatch(fmt, '(.)') do
				local v = nil
				if c == 'i' or c == 'h' or c == 'r' then
					v, offset = string.match(str, '%s*([^%s]+)()', offset)
					v = tonumber(v)
				elseif c == 'f' or c == 'd' or t == 't'  then
					v, offset = string.match(str, '%s*([^%s]+)()', offset)
					v = tonumber(v)
				elseif c == 's' or c == 'S' then
					v, offset  = string.match(str, '%s*(%b\'\')()', offset)
					v = string.sub(v, 2, #v-1) -- strip single quotes
				elseif c == 'b' then
					--FIXME
				elseif c == 'm' then
					--FIXME
				elseif c == 'c' then
					v, offset  = string.match(str, '%s*(%a)()', offset)
					v = string.byte(v)
				elseif c == 'T' then
					-- nothing
				elseif c == 'F' then
					-- nothing
				elseif c == 'N' then
					-- nothing
				elseif c == 'I' then
					-- nothing
				end

				if v then
					table.insert(tab, v)
				end
			end

			forge:message(path, fmt, table.unpack(tab))
		else
			forge:message(path)
		end
	end
end

local osc_lookup = {
	[ATOM.Int] = 'i',
	[ATOM.Long] = 'h',
	[ATOM.Float] = 'f',
	[ATOM.Double] = 'd',
	[ATOM.Bool] = function(atom) return atom.body and 'T' or 'F' end,
	[ATOM.URID] = 'S',
	[ATOM.String] = 's',
	[ATOM.Literal] = function(atom)
		if atom.datatype == OSC.Nil then return 'N'
		elseif atom.datatype == OSC.Impulse then return 'I'
		elseif atom.datatype == OSC.Char then return 'c'
		elseif atom.datatype == OSC.RGBA then return 'r'
		end
	end,
	[ATOM.Chunk] = 'b',
	[ATOM.Object] = function(atom) return atom.otype == OSC.Timestamp and 't' end,
	[MIDI.MidiEvent] = 'm'
}

local function osc_tostring(atom)
	return tostring(atom.body)
end

local function osc_tochunk(atom)
	local res = '(' .. #atom ')'

	local body = atom.body
	for i = 1, #atom do
		res = res .. string.format('%02X', string.byte(body, i))
	end

	return res
end

local osc_echo = {
	[ATOM.Int] = osc_tostring,
	[ATOM.Long] = osc_tostring,
	[ATOM.Float] = osc_tostring,
	[ATOM.Double] = osc_tostring,
	[ATOM.Bool] = nil,
	[ATOM.URID] = function(atom) return "'" .. Unmap(atom.body) .. "'" end,
	[ATOM.String] = function(atom) return "'" .. atom.body .. "'" end,
	[ATOM.Literal] = function(atom)
		if atom.datatype == OSC.Char then return atom.body
		elseif atom.datatype == OSC.RGBA then return '0x' .. atom.body end
	end,
	[ATOM.Chunk] = osc_tochunk,
	[ATOM.Object] = function(atom)
		local integral = atom[OSC.timestampIntegral].body
		local fraction = atom[OSC.timestampFraction].body
		return string.format('%08x.%08x', integral, fraction)
	end,
	[MIDI.MidiEvent] = osc_tochunk
}

local function osc_print(obj)
	local path = obj[OSC.messagePath].body
	local args = obj[OSC.messageArguments]

	local res = path
	local fmt = ','

	if args and (#args > 0) then
		for k, atom in args:foreach() do
			local lup = osc_lookup[atom.type]
			if lup then
				if type(lup) == 'function' then
					lup = lup(atom)
				end

				fmt = fmt .. lup
			end
		end

		res = res .. ' ' .. fmt

		for k, atom in args:foreach() do
			local ech = osc_echo[atom.type]
			if ech then
				local tmp = ech(atom)
				if tmp then
					res = res .. ' ' .. tmp
				end
			end
		end
	end

	return res, midi_col[0x0], true
end

local function osc_edit(atom)
	if not atom or atom.type ~= ATOM.Object or atom.otype ~= OSC.Message then
		return ''
	end

	return osc_print(atom)
end

local function osc_write(atom)
	if not atom then
		return '...', midi_col[0x0], false
	elseif atom.type ~= ATOM.Object or atom.otype ~= OSC.Message then
		return Unmap(atom.otype or atom.type), midi_col[0x0], false
	end

	return osc_print(atom)
end

local osc_mode = {
	label = 'OSC',
	parse= osc_parse,
	edit= osc_edit,
	write = osc_write
}

_P = {
	midi_dec_mode,
	midi_hex_mode,
	osc_mode
}
